@@ -11,6 +11,7 @@ use serde_crate::de::{self, Deserialize, Visitor};
 use serde_crate::de::IntoDeserializer;
 
 use webcore::value::{
+    self,
     Undefined,
     Null,
     Value
@@ -155,12 +156,16 @@ impl Serialize for Value {
             Value::Bool( value ) => serializer.serialize_bool( value ),
             Value::Number( ref value ) => value.serialize( serializer ),
             Value::String( ref value ) => serializer.serialize_str( value ),
-            Value::Array( ref value ) => value.serialize( serializer ),
+            Value::Array( ref value ) => {
+                let value: Vec< Value > = value.into();
+                value.serialize( serializer )
+            },
             Value::Object( ref value ) => {
+                let value: BTreeMap< String, Value > = value.into();
                 let mut map = try!( serializer.serialize_map( Some( value.len() ) ) );
                 for (key, value) in value {
-                    try!( map.serialize_key( key ) );
-                    try!( map.serialize_value( value ) );
+                    try!( map.serialize_key( &key ) );
+                    try!( map.serialize_value( &value ) );
                 }
 
                 map.end()
@@ -356,6 +361,16 @@ impl From< number::ConversionError > for ConversionError {
     }
 }
 
+impl From< ConversionError > for value::ConversionError {
+    fn from( error: ConversionError ) -> Self {
+        match error.kind {
+            ConversionErrorKind::InvalidKey => value::ConversionError::Custom( "key must be either a string or an integer".to_owned() ),
+            ConversionErrorKind::NumberConversionError( error ) => error.into(),
+            ConversionErrorKind::Custom( message ) => value::ConversionError::Custom( message )
+        }
+    }
+}
+
 #[derive(Debug)]
 pub struct Serializer {
 }
@@ -460,7 +475,7 @@ impl< 'a > ser::Serializer for &'a mut Serializer {
     fn serialize_newtype_variant< T: ?Sized + Serialize >( self, _name: &'static str, _variant_index: u32, variant: &'static str, value: &T ) -> Result< Self::Ok, Self::Error > {
         let mut object = BTreeMap::new();
         object.insert( String::from( variant ), to_value( &value )? );
-        Ok( Value::Object( object ) )
+        Ok( Value::Object( object.into() ) )
     }
 
     fn serialize_seq( self, length: Option< usize > ) -> Result< Self::SerializeSeq, Self::Error > {
@@ -599,8 +614,8 @@ impl ser::SerializeTupleVariant for SerializeTupleVariant {
 
     fn end( self ) -> Result< Self::Ok, Self::Error > {
         let mut object: BTreeMap< String, Value > = BTreeMap::new();
-        object.insert( self.name, Value::Array( self.elements ) );
-        Ok( Value::Object( object ) )
+        object.insert( self.name, Value::Array( self.elements.into() ) );
+        Ok( Value::Object( object.into() ) )
     }
 }
 
@@ -638,7 +653,7 @@ impl ser::SerializeMap for SerializeMap {
     }
 
     fn end( self ) -> Result< Self::Ok, Self::Error > {
-        Ok( Value::Object( self.map ) )
+        Ok( Value::Object( self.map.into() ) )
     }
 }
 
@@ -667,8 +682,8 @@ impl ser::SerializeStructVariant for SerializeStructVariant {
 
     fn end( self ) -> Result< Self::Ok, Self::Error > {
         let mut object = BTreeMap::new();
-        object.insert( self.name, Value::Object( self.map ) );
-        Ok( Value::Object( object ) )
+        object.insert( self.name, Value::Object( self.map.into() ) );
+        Ok( Value::Object( object.into() ) )
     }
 }
 
@@ -725,6 +740,7 @@ impl< 'de > de::Deserializer< 'de > for Value {
             Value::Number( value ) => de::Deserializer::deserialize_any( value, visitor ),
             Value::String( value ) => visitor.visit_string( value ),
             Value::Array( value ) => {
+                let value: Vec< _ > = value.into();
                 let length = value.len();
                 let mut deserializer = SeqDeserializer::new( value );
                 let seq = visitor.visit_seq( &mut deserializer )?;
@@ -736,6 +752,7 @@ impl< 'de > de::Deserializer< 'de > for Value {
                 }
             },
             Value::Object( value ) => {
+                let value: BTreeMap< _, _ > = value.into();
                 let length = value.len();
                 let mut deserializer = MapDeserializer::new( value );
                 let map = visitor.visit_map( &mut deserializer )?;
@@ -765,6 +782,7 @@ impl< 'de > de::Deserializer< 'de > for Value {
     fn deserialize_enum< V: Visitor< 'de > >( self, _name: &str, _variants: &'static [&'static str], visitor: V ) -> Result< V::Value, Self::Error > {
         let (variant, value) = match self {
             Value::Object( value ) => {
+                let value: BTreeMap< _, _ > = value.into();
                 let mut iter = value.into_iter();
                 let (variant, value) = match iter.next() {
                     Some( value ) => value,
@@ -844,7 +862,7 @@ impl< 'de > de::VariantAccess< 'de > for VariantDeserializer {
     fn tuple_variant< V: Visitor< 'de > >( self, _length: usize, visitor: V ) -> Result< V::Value, Self::Error > {
         match self.value {
             Some( Value::Array( value ) ) => {
-                de::Deserializer::deserialize_any( SeqDeserializer::new( value ), visitor )
+                de::Deserializer::deserialize_any( SeqDeserializer::new( value.into() ), visitor )
             },
             Some( other ) => Err( de::Error::invalid_type( other.unexpected(), &"tuple variant" ) ),
             None => Err( de::Error::invalid_type( de::Unexpected::UnitVariant, &"tuple variant" ) )
@@ -854,7 +872,7 @@ impl< 'de > de::VariantAccess< 'de > for VariantDeserializer {
     fn struct_variant< V: Visitor< 'de > >( self, _fields: &'static [&'static str], visitor: V ) -> Result< V::Value, Self::Error > {
         match self.value {
             Some( Value::Object( value ) ) => {
-                de::Deserializer::deserialize_any( MapDeserializer::new( value ), visitor )
+                de::Deserializer::deserialize_any( MapDeserializer::new( value.into() ), visitor )
             },
             Some( other ) => Err( de::Error::invalid_type( other.unexpected(), &"struct variant" ) ),
             _ => Err( de::Error::invalid_type( de::Unexpected::UnitVariant, &"struct variant" ) )
@@ -993,7 +1011,7 @@ macro_rules! __js_serializable_serde_boilerplate {
             fn memory_required( &self ) -> usize {
                 // TODO: This is very inefficient. The actual conversion into
                 // the Value should be only done once.
-                let value = to_value( self ).unwrap();
+                let value = $crate::private::to_value( self ).unwrap();
                 $crate::private::JsSerializable::memory_required( &value )
             }
         }
@@ -1347,7 +1365,10 @@ mod tests {
     #[test]
     fn deserialize_value_array() {
         let value: Value = serde_json::from_str( "[true, false]" ).unwrap();
-        assert_eq!( value, Value::Array( vec![ Value::Bool( true ), Value::Bool( false ) ] ) );
+        assert_eq!( value.is_array(), true );
+
+        let value: Vec< Value > = value.try_into().unwrap();
+        assert_eq!( value, vec![ Value::Bool( true ), Value::Bool( false ) ] );
     }
 
     #[test]
@@ -1356,7 +1377,9 @@ mod tests {
         let mut map: BTreeMap< String, Value > = BTreeMap::new();
         map.insert( "1".to_owned(), Value::String( "one".to_owned() ) );
         map.insert( "2".to_owned(), Value::String( "two".to_owned() ) );
-        assert_eq!( value, Value::Object( map ) );
+
+        let value: BTreeMap< _, _ > = value.into_object().unwrap().into();
+        assert_eq!( value, map );
     }
 
     #[derive(Serialize, Deserialize, Debug)]
@@ -1365,7 +1388,7 @@ mod tests {
         string: String
     }
 
-    #[derive(Serialize, Deserialize, Debug)]
+    #[derive(PartialEq, Serialize, Deserialize, Debug)]
     struct StructureSerializable {
         number: i32,
         string: String
@@ -1385,7 +1408,9 @@ mod tests {
         let mut map = BTreeMap::new();
         map.insert( "number".to_owned(), Value::Number( 123.into() ) );
         map.insert( "string".to_owned(), Value::String( "Hello!".to_owned() ) );
-        assert_eq!( value, Value::Object( map ) );
+
+        let value: BTreeMap< _, _ > = value.into_object().unwrap().into();
+        assert_eq!( value, map );
     }
 
     #[test]
@@ -1414,7 +1439,9 @@ mod tests {
         let mut map = BTreeMap::new();
         map.insert( "number".to_owned(), Value::Number( 123.into() ) );
         map.insert( "string".to_owned(), Value::String( "Hello!".to_owned() ) );
-        assert_eq!( value, Value::Object( map ) );
+
+        let value: BTreeMap< _, _ > = value.into_object().unwrap().into();
+        assert_eq!( value, map );
     }
 
     #[test]
@@ -1459,4 +1486,33 @@ mod tests {
         assert_eq!( structure.0.number, 123 );
         assert_eq!( structure.0.string, "Hello!" );
     }
+
+    #[test]
+    fn serialization_and_deserialization_of_btreemap_with_serializable_values() {
+        let mut original = BTreeMap::new();
+        original.insert( "key", StructureSerializable {
+            number: 123,
+            string: "Hello!".to_owned()
+        });
+
+        let value: Value = (&original).into();
+        let deserialized: BTreeMap< String, StructureSerializable > = value.try_into().unwrap();
+
+        assert_eq!( original.len(), deserialized.len() );
+        assert_eq!( original.get( "key" ).unwrap(), deserialized.get( "key" ).unwrap() );
+    }
+
+    #[test]
+    fn serialization_and_deserialization_of_array_with_serializable_elements() {
+        let original = vec![ StructureSerializable {
+            number: 123,
+            string: "Hello!".to_owned()
+        }];
+
+        let value: Value = (&original).into();
+        let deserialized: Vec< StructureSerializable > = value.try_into().unwrap();
+
+        assert_eq!( original.len(), deserialized.len() );
+        assert_eq!( original[ 0 ], deserialized[ 0 ] );
+    }
 }
