@@ -70,7 +70,7 @@
 #![cfg_attr(feature = "dev", allow(unstable_features))]
 #![cfg_attr(feature = "dev", feature(plugin))]
 #![cfg_attr(feature = "dev", plugin(clippy))]
-#![recursion_limit="750"]
+#![recursion_limit="1500"]
 
 #[cfg(feature = "serde")]
 #[macro_use]
@@ -99,6 +99,8 @@ pub use webcore::value::{
     Reference
 };
 pub use webcore::number::Number;
+pub use webcore::object::Object;
+pub use webcore::array::Array;
 
 #[cfg(feature = "serde")]
 /// A module with serde-related APIs.
@@ -123,8 +125,9 @@ pub mod web {
         set_timeout,
         alert
     };
+    pub use webapi::cross_origin_setting::CrossOriginSetting;
     pub use webapi::date::Date;
-    pub use webapi::event_target::{IEventTarget, EventTarget};
+    pub use webapi::event_target::{IEventTarget, EventTarget, EventListenerHandle};
     pub use webapi::node::{INode, Node, CloneKind};
     pub use webapi::element::{IElement, Element};
     pub use webapi::html_element::{IHtmlElement, HtmlElement};
@@ -144,12 +147,15 @@ pub mod web {
 
     /// A module containing HTML DOM elements.
     pub mod html_element {
+        pub use webapi::html_elements::ImageElement;
         pub use webapi::html_elements::InputElement;
     }
 
     /// A module containing JavaScript DOM events.
     pub mod event {
         pub use webapi::event::{
+            ConcreteEvent,
+
             IEvent,
             IKeyboardEvent,
             IUiEvent,
@@ -161,7 +167,7 @@ pub mod web {
             KeyboardLocation,
             ModifierKey,
             MouseButton,
-            
+
             ChangeEvent,
             KeypressEvent,
             ClickEvent,
@@ -192,7 +198,7 @@ pub mod unstable {
 
 #[doc(hidden)]
 pub mod private {
-    pub use webcore::ffi::emscripten_asm_const_int;
+    pub use webcore::ffi::exports::*;
     pub use webcore::serialization::{
         JsSerializable,
         JsSerializableOwned,
