@@ -4,6 +4,7 @@ use webcore::value::Reference;
 use webcore::try_from::TryInto;
 use webapi::event::{ConcreteEvent, IEvent};
 
+/// A handle to a particular event listener.
 pub struct EventListenerHandle {
     event_type: &'static str,
     reference: Reference,
@@ -47,12 +48,12 @@ pub trait IEventTarget: AsRef< Reference > {
         let reference = self.as_ref();
         let listener_reference = js! {
             var listener = @{listener};
-            @{reference}.addEventListener( @{T::static_event_type()}, listener );
+            @{reference}.addEventListener( @{T::EVENT_TYPE}, listener );
             return listener;
         }.try_into().unwrap();
 
         EventListenerHandle {
-            event_type: T::static_event_type(),
+            event_type: T::EVENT_TYPE,
             reference: reference.clone(),
             listener_reference: listener_reference
         }
