@@ -1,21 +1,25 @@
 use std::mem;
 use std::slice;
 use std::i32;
-use std::collections::BTreeMap;
+use std::collections::{BTreeMap, HashMap};
 use std::marker::PhantomData;
 use std::cell::{Cell, UnsafeCell};
+use std::hash::Hash;
 
 use webcore::ffi;
 use webcore::callfn::CallMut;
 use webcore::newtype::Newtype;
 use webcore::try_from::{TryFrom, TryInto};
 use webcore::number::Number;
+use webcore::object::Object;
+use webcore::array::Array;
 
 use webcore::value::{
     Null,
     Undefined,
     Reference,
-    Value
+    Value,
+    FromReferenceUnchecked
 };
 
 #[repr(u8)]
@@ -31,7 +35,9 @@ pub enum Tag {
     Array = 7,
     Object = 8,
     Reference = 9,
-    Function = 10
+    Function = 10,
+    ObjectReference = 11,
+    ArrayReference = 12
 }
 
 impl Default for Tag {
@@ -71,7 +77,7 @@ impl PreallocatedArena {
     pub fn reserve< T >( &self, count: usize ) -> &mut [T] {
         let bytes = mem::size_of::< T >() * count;
         let slice = unsafe {
-            let mut buffer = &mut *self.buffer.get();
+            let buffer = &mut *self.buffer.get();
             debug_assert!( self.index.get() + bytes <= buffer.len() );
 
             slice::from_raw_parts_mut(
@@ -88,7 +94,7 @@ impl PreallocatedArena {
     #[inline]
     pub fn save( &self, value: Value ) -> &Value {
         unsafe {
-            let mut saved = &mut *self.saved.get();
+            let saved = &mut *self.saved.get();
             saved.push( value );
             &*(saved.last().unwrap() as *const Value)
         }
@@ -193,12 +199,28 @@ struct SerializedUntaggedFunction {
     deallocator_pointer: u32
 }
 
+#[repr(C)]
+#[derive(Debug)]
+struct SerializedUntaggedObjectReference {
+    refid: i32
+}
+
+#[repr(C)]
+#[derive(Debug)]
+struct SerializedUntaggedArrayReference {
+    refid: i32
+}
+
 impl SerializedUntaggedString {
     #[inline]
     fn deserialize( &self ) -> String {
         let pointer = self.pointer as *mut u8;
         let length = self.length as usize;
 
+        if length == 0 {
+            return String::new();
+        }
+
         unsafe {
             let vector = Vec::from_raw_parts( pointer, length, length + 1 );
             String::from_utf8_unchecked( vector )
@@ -217,36 +239,132 @@ impl SerializedUntaggedArray {
 
         let vector = slice.iter().map( |value| value.deserialize() ).collect();
         unsafe {
-            ffi::free( pointer as *const u8 );
+            ffi::dealloc( pointer as *mut u8, length * mem::size_of::< SerializedValue >() );
         }
 
         vector
     }
 }
 
-impl SerializedUntaggedObject {
-    #[inline]
-    fn deserialize( &self ) -> BTreeMap< String, Value > {
-        let length = self.length as usize;
-        let key_pointer = self.key_pointer as *const SerializedUntaggedString;
-        let value_pointer = self.value_pointer as *const SerializedValue;
+pub struct ObjectDeserializer< 'a > {
+    key_slice: &'a [SerializedUntaggedString],
+    value_slice: &'a [SerializedValue< 'a >],
+    index: usize
+}
 
-        let key_slice = unsafe { slice::from_raw_parts( key_pointer, length ) };
-        let value_slice = unsafe { slice::from_raw_parts( value_pointer, length ) };
-
-        let map = key_slice.iter().zip( value_slice.iter() ).map( |(key, value)| {
-            let key = key.deserialize();
-            let value = value.deserialize();
-            (key, value)
-        }).collect();
-
-        unsafe {
-            ffi::free( key_pointer as *const u8 );
-            ffi::free( value_pointer as *const u8 );
+impl< 'a > Iterator for ObjectDeserializer< 'a > {
+    type Item = (String, Value);
+    fn next( &mut self ) -> Option< Self::Item > {
+        if self.index >= self.key_slice.len() {
+            None
+        } else {
+            let key = self.key_slice[ self.index ].deserialize();
+            let value = self.value_slice[ self.index ].deserialize();
+            self.index += 1;
+            Some( (key, value) )
         }
-
-        map
     }
+
+    #[inline]
+    fn size_hint( &self ) -> (usize, Option< usize >) {
+        let remaining = self.key_slice.len() - self.index;
+        (remaining, Some( remaining ))
+    }
+}
+
+impl< 'a > ExactSizeIterator for ObjectDeserializer< 'a > {}
+
+pub fn deserialize_object< R, F: FnOnce( &mut ObjectDeserializer ) -> R >( reference: &Reference, callback: F ) -> R {
+    let mut result: SerializedValue = Default::default();
+    __js_raw_asm!( "\
+        var object = Module.STDWEB.acquire_js_reference( $0 );\
+        Module.STDWEB.serialize_object( $1, object );",
+        reference.as_raw(),
+        (&mut result as *mut _)
+    );
+
+    assert_eq!( result.tag, Tag::Object );
+    let result = result.as_object();
+
+    let length = result.length as usize;
+    let key_pointer = result.key_pointer as *const SerializedUntaggedString;
+    let value_pointer = result.value_pointer as *const SerializedValue;
+
+    let key_slice = unsafe { slice::from_raw_parts( key_pointer, length ) };
+    let value_slice = unsafe { slice::from_raw_parts( value_pointer, length ) };
+
+    let mut iter = ObjectDeserializer {
+        key_slice,
+        value_slice,
+        index: 0
+    };
+
+    let output = callback( &mut iter );
+
+    // TODO: Panic-safety.
+    unsafe {
+        ffi::dealloc( key_pointer as *mut u8, length * mem::size_of::< SerializedUntaggedString >() );
+        ffi::dealloc( value_pointer as *mut u8, length * mem::size_of::< SerializedValue >() );
+    }
+
+    output
+}
+
+pub struct ArrayDeserializer< 'a > {
+    slice: &'a [SerializedValue< 'a >],
+    index: usize
+}
+
+impl< 'a > Iterator for ArrayDeserializer< 'a > {
+    type Item = Value;
+    fn next( &mut self ) -> Option< Self::Item > {
+        if self.index >= self.slice.len() {
+            None
+        } else {
+            let value = self.slice[ self.index ].deserialize();
+            self.index += 1;
+            Some( value )
+        }
+    }
+
+    #[inline]
+    fn size_hint( &self ) -> (usize, Option< usize >) {
+        let remaining = self.slice.len() - self.index;
+        (remaining, Some( remaining ))
+    }
+}
+
+impl< 'a > ExactSizeIterator for ArrayDeserializer< 'a > {}
+
+pub fn deserialize_array< R, F: FnOnce( &mut ArrayDeserializer ) -> R >( reference: &Reference, callback: F ) -> R {
+    let mut result: SerializedValue = Default::default();
+    __js_raw_asm!( "\
+        var array = Module.STDWEB.acquire_js_reference( $0 );\
+        Module.STDWEB.serialize_array( $1, array );",
+        reference.as_raw(),
+        (&mut result as *mut _)
+    );
+
+    assert_eq!( result.tag, Tag::Array );
+    let result = result.as_array();
+
+    let length = result.length as usize;
+    let pointer = result.pointer as *const SerializedValue;
+
+    let slice = unsafe { slice::from_raw_parts( pointer, length ) };
+    let mut iter = ArrayDeserializer {
+        slice,
+        index: 0
+    };
+
+    let output = callback( &mut iter );
+
+    // TODO: Panic-safety.
+    unsafe {
+        ffi::dealloc( pointer as *mut u8, length * mem::size_of::< SerializedValue >() );
+    }
+
+    output
 }
 
 impl SerializedUntaggedReference {
@@ -256,6 +374,26 @@ impl SerializedUntaggedReference {
     }
 }
 
+impl SerializedUntaggedObjectReference {
+    #[inline]
+    fn deserialize( &self ) -> Object {
+        unsafe {
+            let reference = Reference::from_raw_unchecked( self.refid );
+            Object::from_reference_unchecked( reference )
+        }
+    }
+}
+
+impl SerializedUntaggedArrayReference {
+    #[inline]
+    fn deserialize( &self ) -> Array {
+        unsafe {
+            let reference = Reference::from_raw_unchecked( self.refid );
+            Array::from_reference_unchecked( reference )
+        }
+    }
+}
+
 macro_rules! untagged_boilerplate {
     ($tests_namespace:ident, $reader_name:ident, $tag:expr, $untagged_type:ident) => {
         impl< 'a > SerializedValue< 'a > {
@@ -306,6 +444,8 @@ untagged_boilerplate!( test_string, as_string, Tag::Str, SerializedUntaggedStrin
 untagged_boilerplate!( test_array, as_array, Tag::Array, SerializedUntaggedArray );
 untagged_boilerplate!( test_reference, as_reference, Tag::Reference, SerializedUntaggedReference );
 untagged_boilerplate!( test_function, as_function, Tag::Function, SerializedUntaggedFunction );
+untagged_boilerplate!( test_object_reference, as_object_reference, Tag::ObjectReference, SerializedUntaggedObjectReference );
+untagged_boilerplate!( test_array_reference, as_array_reference, Tag::ArrayReference, SerializedUntaggedArrayReference );
 
 impl< 'a > SerializedValue< 'a > {
     #[doc(hidden)]
@@ -319,10 +459,10 @@ impl< 'a > SerializedValue< 'a > {
             Tag::Str => Value::String( self.as_string().deserialize() ),
             Tag::False => Value::Bool( false ),
             Tag::True => Value::Bool( true ),
-            Tag::Array => Value::Array( self.as_array().deserialize() ),
-            Tag::Object => Value::Object( self.as_object().deserialize() ),
+            Tag::ObjectReference => Value::Object( self.as_object_reference().deserialize() ),
+            Tag::ArrayReference => Value::Array( self.as_array_reference().deserialize() ),
             Tag::Reference => self.as_reference().deserialize().into(),
-            Tag::Function => unreachable!()
+            Tag::Function | Tag::Object | Tag::Array => unreachable!()
         }
     }
 }
@@ -385,6 +525,38 @@ impl JsSerializable for Reference {
 
 __js_serializable_boilerplate!( Reference );
 
+impl JsSerializable for Object {
+    #[inline]
+    fn into_js< 'a >( &'a self, _: &'a PreallocatedArena ) -> SerializedValue< 'a > {
+        SerializedUntaggedObjectReference {
+            refid: self.as_reference().as_raw()
+        }.into()
+    }
+
+    #[inline]
+    fn memory_required( &self ) -> usize {
+        0
+    }
+}
+
+__js_serializable_boilerplate!( Object );
+
+impl JsSerializable for Array {
+    #[inline]
+    fn into_js< 'a >( &'a self, _: &'a PreallocatedArena ) -> SerializedValue< 'a > {
+        SerializedUntaggedArrayReference {
+            refid: self.as_reference().as_raw()
+        }.into()
+    }
+
+    #[inline]
+    fn memory_required( &self ) -> usize {
+        0
+    }
+}
+
+__js_serializable_boilerplate!( Array );
+
 impl JsSerializable for bool {
     #[inline]
     fn into_js< 'a >( &'a self, _: &'a PreallocatedArena ) -> SerializedValue< 'a > {
@@ -625,32 +797,54 @@ impl< T: JsSerializable > JsSerializable for [T] {
 
 __js_serializable_boilerplate!( impl< 'a, T > for &'a [T] where T: JsSerializable );
 
-impl< T: JsSerializable > JsSerializable for BTreeMap< String, T > {
+fn object_into_js< 'a, K: AsRef< str >, V: 'a + JsSerializable, I: Iterator< Item = (K, &'a V) > + ExactSizeIterator >( iter: I, arena: &'a PreallocatedArena ) -> SerializedValue< 'a > {
+    let keys = arena.reserve( iter.len() );
+    let values = arena.reserve( iter.len() );
+    for (((key, value), output_key), output_value) in iter.zip( keys.iter_mut() ).zip( values.iter_mut() ) {
+        *output_key = key.as_ref().into_js( arena ).as_string().clone();
+        *output_value = value.into_js( arena );
+    }
+
+    SerializedUntaggedObject {
+        key_pointer: keys.as_ptr() as u32,
+        value_pointer: values.as_ptr() as u32,
+        length: keys.len() as u32
+    }.into()
+}
+
+fn object_memory_required< K: AsRef< str >, V: JsSerializable, I: Iterator< Item = (K, V) > + ExactSizeIterator >( iter: I ) -> usize {
+    mem::size_of::< SerializedValue >() * iter.len() +
+    mem::size_of::< SerializedUntaggedString >() * iter.len() +
+    iter.fold( 0, |sum, (key, value)| sum + key.as_ref().memory_required() + value.memory_required() )
+}
+
+impl< K: AsRef< str >, V: JsSerializable > JsSerializable for BTreeMap< K, V > {
     #[inline]
     fn into_js< 'a >( &'a self, arena: &'a PreallocatedArena ) -> SerializedValue< 'a > {
-        let keys = arena.reserve( self.len() );
-        let values = arena.reserve( self.len() );
-        for (((key, value), output_key), output_value) in self.iter().zip( keys.iter_mut() ).zip( values.iter_mut() ) {
-            *output_key = key.into_js( arena ).as_string().clone();
-            *output_value = value.into_js( arena );
-        }
-
-        SerializedUntaggedObject {
-            key_pointer: keys.as_ptr() as u32,
-            value_pointer: values.as_ptr() as u32,
-            length: keys.len() as u32
-        }.into()
+        object_into_js( self.iter(), arena )
     }
 
     #[inline]
     fn memory_required( &self ) -> usize {
-        mem::size_of::< SerializedValue >() * self.len() +
-        mem::size_of::< SerializedUntaggedString >() * self.len() +
-        self.iter().fold( 0, |sum, (key, value)| sum + key.memory_required() + value.memory_required() )
+        object_memory_required( self.iter() )
     }
 }
 
-__js_serializable_boilerplate!( impl< T > for BTreeMap< String, T > where T: JsSerializable );
+__js_serializable_boilerplate!( impl< K, V > for BTreeMap< K, V > where K: AsRef< str >, V: JsSerializable );
+
+impl< K: AsRef< str > + Eq + Hash, V: JsSerializable > JsSerializable for HashMap< K, V > {
+    #[inline]
+    fn into_js< 'a >( &'a self, arena: &'a PreallocatedArena ) -> SerializedValue< 'a > {
+        object_into_js( self.iter(), arena )
+    }
+
+    #[inline]
+    fn memory_required( &self ) -> usize {
+        object_memory_required( self.iter() )
+    }
+}
+
+__js_serializable_boilerplate!( impl< K, V > for HashMap< K, V > where K: AsRef< str > + Eq + Hash, V: JsSerializable );
 
 impl JsSerializable for Value {
     fn into_js< 'a >( &'a self, arena: &'a PreallocatedArena ) -> SerializedValue< 'a > {
@@ -660,7 +854,7 @@ impl JsSerializable for Value {
             Value::Bool( ref value ) => value.into_js( arena ),
             Value::Number( ref value ) => value.into_js( arena ),
             Value::String( ref value ) => value.into_js( arena ),
-            Value::Array( ref value ) => value.as_slice().into_js( arena ),
+            Value::Array( ref value ) => value.into_js( arena ),
             Value::Object( ref value ) => value.into_js( arena ),
             Value::Reference( ref value ) => value.into_js( arena )
         }
@@ -673,7 +867,7 @@ impl JsSerializable for Value {
             Value::Bool( value ) => value.memory_required(),
             Value::Number( value ) => value.memory_required(),
             Value::String( ref value ) => value.memory_required(),
-            Value::Array( ref value ) => value.as_slice().memory_required(),
+            Value::Array( ref value ) => value.memory_required(),
             Value::Object( ref value ) => value.memory_required(),
             Value::Reference( ref value ) => value.memory_required()
         }
@@ -682,72 +876,98 @@ impl JsSerializable for Value {
 
 __js_serializable_boilerplate!( Value );
 
+#[derive(Debug)]
+pub struct FunctionTag;
+
+#[derive(Debug)]
+pub struct NonFunctionTag;
+
+impl< T: JsSerializable > JsSerializableOwned for Newtype< (NonFunctionTag, ()), T > {
+    #[inline]
+    fn into_js_owned< 'x >( value: &'x mut Option< Self >, arena: &'x PreallocatedArena ) -> SerializedValue< 'x > {
+        JsSerializable::into_js( value.as_ref().unwrap().as_ref(), arena )
+    }
+
+    #[inline]
+    fn memory_required_owned( &self ) -> usize {
+        JsSerializable::memory_required( &**self )
+    }
+}
+
+trait FuncallAdapter< F > {
+    extern fn funcall_adapter( callback: *mut F, raw_arguments: *mut SerializedUntaggedArray );
+    extern fn deallocator( callback: *mut F );
+}
+
 macro_rules! impl_for_fn {
     ($next:tt => $($kind:ident),*) => {
+        impl< $($kind: TryFrom< Value >,)* F > FuncallAdapter< F > for Newtype< (FunctionTag, ($($kind,)*)), F >
+            where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializableOwned
+        {
+            #[allow(unused_mut, unused_variables, non_snake_case)]
+            extern fn funcall_adapter(
+                    callback: *mut F,
+                    raw_arguments: *mut SerializedUntaggedArray
+                )
+            {
+                let callback = unsafe { &mut *callback };
+                let mut arguments = unsafe { &*raw_arguments }.deserialize();
 
-        impl< $($kind: TryFrom< Value >,)* F > JsSerializableOwned for Newtype< ($($kind,)*), F >
+                unsafe {
+                    ffi::dealloc( raw_arguments as *mut u8, mem::size_of::< SerializedValue >() );
+                }
+
+                if arguments.len() != F::expected_argument_count() {
+                    // TODO: Should probably throw an exception into the JS world or something like that.
+                    panic!( "Expected {} arguments, got {}", F::expected_argument_count(), arguments.len() );
+                }
+
+                let mut arguments = arguments.drain( .. );
+                let mut nth_argument = 0;
+                $(
+                    let $kind = match arguments.next().unwrap().try_into() {
+                        Ok( value ) => value,
+                        Err( _ ) => {
+                            panic!( "Argument #{} is not convertible", nth_argument + 1 );
+                        }
+                    };
+
+                    nth_argument += 1;
+                )*
+
+                $crate::private::noop( &mut nth_argument );
+
+                let result = callback.call_mut( ($($kind,)*) );
+                let mut arena = PreallocatedArena::new( result.memory_required_owned() );
+
+                let mut result = Some( result );
+                let result = JsSerializableOwned::into_js_owned( &mut result, &mut arena );
+                let result = &result as *const _;
+
+                // This is kinda hacky but I'm not sure how else to do it at the moment.
+                __js_raw_asm!( "Module.STDWEB.tmp = Module.STDWEB.to_js( $0 );", result );
+            }
+
+            extern fn deallocator( callback: *mut F ) {
+                let callback = unsafe {
+                    Box::from_raw( callback )
+                };
+
+                drop( callback );
+            }
+        }
+
+        impl< $($kind: TryFrom< Value >,)* F > JsSerializableOwned for Newtype< (FunctionTag, ($($kind,)*)), F >
             where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializableOwned
         {
             #[inline]
-            #[allow(unused_mut, unused_variables, non_snake_case)]
             fn into_js_owned< 'a >( value: &'a mut Option< Self >, _: &'a PreallocatedArena ) -> SerializedValue< 'a > {
                 let callback: *mut F = Box::into_raw( Box::new( value.take().unwrap().unwrap_newtype() ) );
 
-                extern fn funcall_adapter< $($kind: TryFrom< Value >,)* F: CallMut< ($($kind,)*) > >(
-                        callback: *mut F,
-                        raw_arguments: *mut SerializedUntaggedArray
-                    ) where F::Output: JsSerializableOwned
-                {
-                    let callback = unsafe { &mut *callback };
-                    let mut arguments = unsafe { &*raw_arguments }.deserialize();
-
-                    unsafe {
-                        ffi::free( raw_arguments as *const u8 );
-                    }
-
-                    if arguments.len() != F::expected_argument_count() {
-                        // TODO: Should probably throw an exception into the JS world or something like that.
-                        panic!( "Expected {} arguments, got {}", F::expected_argument_count(), arguments.len() );
-                    }
-
-                    let mut arguments = arguments.drain( .. );
-                    let mut nth_argument = 0;
-                    $(
-                        let $kind = match arguments.next().unwrap().try_into() {
-                            Ok( value ) => value,
-                            Err( _ ) => {
-                                panic!( "Argument #{} is not convertible", nth_argument + 1 );
-                            }
-                        };
-
-                        nth_argument += 1;
-                    )*
-
-                    $crate::private::noop( &mut nth_argument );
-
-                    let result = callback.call_mut( ($($kind,)*) );
-                    let mut arena = PreallocatedArena::new( result.memory_required_owned() );
-
-                    let mut result = Some( result );
-                    let result = JsSerializableOwned::into_js_owned( &mut result, &mut arena );
-                    let result = &result as *const _;
-
-                    // This is kinda hacky but I'm not sure how else to do it at the moment.
-                    em_asm_int!( "Module.STDWEB.tmp = Module.STDWEB.to_js( $0 );", result );
-                }
-
-                extern fn deallocator< $($kind: TryFrom< Value >,)* F: CallMut< ($($kind,)*) > >( callback: *mut F ) {
-                    let callback = unsafe {
-                        Box::from_raw( callback )
-                    };
-
-                    drop( callback );
-                }
-
                 SerializedUntaggedFunction {
-                    adapter_pointer: funcall_adapter::< $($kind,)* F > as u32,
+                    adapter_pointer: <Self as FuncallAdapter< F > >::funcall_adapter as u32,
                     pointer: callback as u32,
-                    deallocator_pointer: deallocator::< $($kind,)* F > as u32
+                    deallocator_pointer: <Self as FuncallAdapter< F > >::deallocator as u32
                 }.into()
             }
 
@@ -757,6 +977,29 @@ macro_rules! impl_for_fn {
             }
         }
 
+        impl< $($kind: TryFrom< Value >,)* F > JsSerializableOwned for Newtype< (FunctionTag, ($($kind,)*)), Option< F > >
+            where F: CallMut< ($($kind,)*) > + 'static, F::Output: JsSerializableOwned
+        {
+            #[inline]
+            fn into_js_owned< 'a >( value: &'a mut Option< Self >, _: &'a PreallocatedArena ) -> SerializedValue< 'a > {
+                if let Some( value ) = value.take().unwrap().unwrap_newtype() {
+                    let callback: *mut F = Box::into_raw( Box::new( value ) );
+                    SerializedUntaggedFunction {
+                        adapter_pointer: <Newtype< (FunctionTag, ($($kind,)*)), F > as FuncallAdapter< F > >::funcall_adapter as u32,
+                        pointer: callback as u32,
+                        deallocator_pointer: <Newtype< (FunctionTag, ($($kind,)*)), F > as FuncallAdapter< F > >::deallocator as u32
+                    }.into()
+                } else {
+                    SerializedUntaggedNull.into()
+                }
+            }
+
+            #[inline]
+            fn memory_required_owned( &self ) -> usize {
+                0
+            }
+        }
+
         next! { $next }
     }
 }
@@ -816,14 +1059,49 @@ mod test_deserialization {
         assert_eq!( js! { return "Dog"; }, Value::String( "Dog".to_string() ) );
     }
 
+    #[test]
+    fn empty_string() {
+        assert_eq!( js! { return ""; }, Value::String( "".to_string() ) );
+    }
+
     #[test]
     fn array() {
-        assert_eq!( js! { return [1, 2]; }, Value::Array( vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ] ) );
+        assert_eq!( js! { return [1, 2]; }.is_array(), true );
     }
 
     #[test]
     fn object() {
-        assert_eq!( js! { return {"one": 1, "two": 2}; }, Value::Object( [("one".to_string(), Value::Number(1.into())), ("two".to_string(), Value::Number(2.into()))].iter().cloned().collect() ) );
+        assert_eq!( js! { return {"one": 1, "two": 2}; }.is_object(), true );
+    }
+
+    #[test]
+    fn object_into_btreemap() {
+        let object = js! { return {"one": 1, "two": 2}; }.into_object().unwrap();
+        let object: BTreeMap< String, Value > = object.into();
+        assert_eq!( object, [
+            ("one".to_string(), Value::Number(1.into())),
+            ("two".to_string(), Value::Number(2.into()))
+        ].iter().cloned().collect() );
+    }
+
+    #[test]
+    fn object_into_hashmap() {
+        let object = js! { return {"one": 1, "two": 2}; }.into_object().unwrap();
+        let object: HashMap< String, Value > = object.into();
+        assert_eq!( object, [
+            ("one".to_string(), Value::Number(1.into())),
+            ("two".to_string(), Value::Number(2.into()))
+        ].iter().cloned().collect() );
+    }
+
+    #[test]
+    fn array_into_vector() {
+        let array = js! { return ["one", 1]; }.into_array().unwrap();
+        let array: Vec< Value > = array.into();
+        assert_eq!( array, &[
+            Value::String( "one".to_string() ),
+            Value::Number( 1.into() )
+        ]);
     }
 
     #[test]
@@ -839,7 +1117,10 @@ mod test_deserialization {
             })( 1, 2 );
         };
 
-        assert_eq!( value, Value::Array( vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ] ) );
+        assert_eq!( value.is_array(), true );
+
+        let value: Vec< Value > = value.try_into().unwrap();
+        assert_eq!( value, vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ] );
     }
 
     #[test]
@@ -882,14 +1163,43 @@ mod test_deserialization {
 
         assert_eq!( value.get(), true );
     }
+
+    #[test]
+    fn function_inside_an_option() {
+        let value = Rc::new( Cell::new( 0 ) );
+        let fn_value = value.clone();
+
+        js! {
+            var callback = @{Some( move || { fn_value.set( 1 ); } )};
+            callback();
+            callback.drop();
+        };
+
+        assert_eq!( value.get(), 1 );
+    }
+
+    #[test]
+    #[allow(unused_assignments)]
+    fn function_inside_an_empty_option() {
+        let mut callback = Some( move || () );
+        callback = None;
+
+        let result = js! {
+            var callback = @{callback};
+            return callback === null;
+        };
+
+        assert_eq!( result, Value::Bool( true ) );
+    }
 }
 
 #[cfg(test)]
 mod test_serialization {
     use super::*;
+    use std::borrow::Cow;
 
     #[test]
-    fn object() {
+    fn object_from_btreemap() {
         let object: BTreeMap< _, _ > = [
             ("number".to_string(), Value::Number( 123.into() )),
             ("string".to_string(), Value::String( "Hello!".into() ))
@@ -902,6 +1212,48 @@ mod test_serialization {
         assert_eq!( result, Value::Bool( true ) );
     }
 
+    #[test]
+    fn object_from_borrowed_btreemap() {
+        let object: BTreeMap< _, _ > = [
+            ("number".to_string(), Value::Number( 123.into() ))
+        ].iter().cloned().collect();
+
+        let result = js! {
+            var object = @{&object};
+            return object.number === 123 && Object.keys( object ).length == 1;
+        };
+        assert_eq!( result, Value::Bool( true ) );
+    }
+
+    #[test]
+    fn object_from_btreemap_with_convertible_key_and_value() {
+        let key: Cow< str > = "number".into();
+        let object: BTreeMap< _, _ > = [
+            (key, 123)
+        ].iter().cloned().collect();
+
+        let result = js! {
+            var object = @{object};
+            return object.number === 123 && Object.keys( object ).length == 1;
+        };
+        assert_eq!( result, Value::Bool( true ) );
+    }
+
+    #[test]
+    fn object_from_hashmap() {
+        let object: HashMap< _, _ > = [
+            ("number".to_string(), Value::Number( 123.into() )),
+            ("string".to_string(), Value::String( "Hello!".into() ))
+        ].iter().cloned().collect();
+
+        let result = js! {
+            var object = @{object};
+            return object.number === 123 && object.string === "Hello!" && Object.keys( object ).length == 2;
+        };
+        assert_eq!( result, Value::Bool( true ) );
+    }
+
+
     #[test]
     fn multiple() {
         let reference: Reference = js! {
@@ -959,16 +1311,32 @@ mod test_reserialization {
         assert_eq!( js! { return @{"Dog"}; }, Value::String( "Dog".to_string() ) );
     }
 
+    #[test]
+    fn empty_string() {
+        assert_eq!( js! { return @{""}; }, Value::String( "".to_string() ) );
+    }
+
     #[test]
     fn array() {
-        let array = vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ];
-        assert_eq!( js! { return @{&[1, 2][..]}; }, Value::Array( array ) );
+        let array: Array = vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ].into();
+        assert_eq!( js! { return @{&array}; }, Value::Array( array ) );
+    }
+
+    #[test]
+    fn array_values_are_not_compared_by_value() {
+        let array: Array = vec![ Value::Number( 1.into() ), Value::Number( 2.into() ) ].into();
+        assert_ne!( js! { return @{&[1, 2][..]}; }, Value::Array( array ) );
     }
 
     #[test]
     fn object() {
-        let object = [("one".to_string(), Value::Number(1.into())), ("two".to_string(), Value::Number(2.into()))].iter().cloned().collect();
-        assert_eq!( js! { return @{&object}; }, Value::Object( object ) );
+        let object: BTreeMap< _, _ > = [
+            ("one".to_string(), Value::Number( 1.into() )),
+            ("two".to_string(), Value::Number( 2.into() ))
+        ].iter().cloned().collect();
+
+        let object: Value = object.into();
+        assert_eq!( js! { return @{&object} }, object );
     }
 
     #[test]
@@ -1013,4 +1381,25 @@ mod test_reserialization {
 
         assert_eq!( value, Value::Number( 0x12345678_i32.into() ) );
     }
+
+    #[test]
+    fn string_identity_function() {
+        fn identity( string: String ) -> String {
+            string
+        }
+
+        let empty = js! {
+            var identity = @{identity};
+            return identity( "" );
+        };
+
+        assert_eq!( empty, Value::String( "".to_string() ) );
+
+        let non_empty = js! {
+            var identity = @{identity};
+            return identity( "死神はりんごしか食べない!" );
+        };
+
+        assert_eq!( non_empty, Value::String( "死神はりんごしか食べない!".to_string() ) );
+    }
 }
